use {
    color_eyre::{
        eyre::{bail, ensure},
        Result,
    },
    std::{cell::Cell, collections::HashMap, io::stdin, rc::Rc},
};

#[derive(Clone, Copy)]
struct SliceCols<'a, T> {
    cols: usize,
    data: &'a [T],
}

fn separate_numbers(matrix: SliceCols<u8>) -> impl Iterator<Item = &[u8]> {
    matrix
        .data
        .chunks(matrix.cols)
        .flat_map(|chunk| chunk.split(|&b| !char::from(b).is_ascii_digit()))
        .filter(|s| !s.is_empty())
}

fn around_indices(ix: usize, cols: usize) -> impl Iterator<Item = usize> {
    let cols_sig = isize::try_from(cols).expect("Length can't be larger than isize::MAX");
    (ix % cols != 0)
        .then_some([-1, cols_sig - 1, -cols_sig - 1])
        .into_iter()
        .chain((ix % cols != cols - 1).then_some([1, cols_sig + 1, -cols_sig + 1]))
        .flatten()
        .chain([-cols_sig, cols_sig])
        .flat_map(move |offset| ix.checked_add_signed(offset))
}

fn part1(matrix: SliceCols<u8>) -> u64 {
    separate_numbers(matrix)
        .map(move |slice| {
            let ix = slice.as_ptr() as usize - matrix.data.as_ptr() as usize;
            (ix..slice.len() + ix)
                .flat_map(|ix| around_indices(ix, matrix.cols))
                .flat_map(|ix| matrix.data.get(ix))
                .any(|&b| (b as char).is_ascii_punctuation() && b != b'.')
                .then(|| {
                    String::from_utf8_lossy(slice)
                        .parse::<u64>()
                        .expect("Filter already isolated numbers")
                })
                .unwrap_or_default()
        })
        .sum()
}

fn part2(matrix: SliceCols<u8>) -> u64 {
    let numbers: HashMap<usize, Rc<Cell<Option<u64>>>> = separate_numbers(matrix)
        .flat_map(|slice| {
            let num = Rc::new(Cell::new(Some(
                String::from_utf8_lossy(slice)
                    .parse::<u64>()
                    .expect("Filter already isolated numbers"),
            )));
            slice
                .iter()
                .map(|i| i as *const u8 as usize - matrix.data.as_ptr() as usize)
                .zip(std::iter::repeat(num))
        })
        .collect();

    matrix
        .data
        .iter()
        .filter(|&&b| b == b'*')
        .map(|i| i as *const u8 as usize - matrix.data.as_ptr() as usize)
        .map(|ix| {
            let mut it = around_indices(ix, matrix.cols).flat_map(|ix| {
                let i = numbers.get(&ix)?;
                i.take()
            });

            it.next().unwrap_or_default() * it.next().unwrap_or_default() * it.product::<u64>()
        })
        .sum()
}

fn main() -> Result<()> {
    let mut lines = stdin().lines();
    let line = match lines.next() {
        Some(Ok(line)) => line,
        Some(Err(e)) => bail!(e),
        None => {
            println!("0");
            return Ok(());
        }
    };

    let cols = line.len();
    let mut matrix = line.as_bytes().to_owned();

    lines.try_for_each(|ln| {
        let ln = ln?;
        ensure!(ln.len() == cols, "Invalid column count");
        matrix.extend_from_slice(ln.as_bytes());
        Ok(())
    })?;

    let matrix = SliceCols {
        cols,
        data: &matrix,
    };

    println!("{}", part2(matrix));

    Ok(())
}
